import os
from pathlib import Path

from dotenv import load_dotenv

load_dotenv()


######################
# cache configurations
######################

CACHE_DIR = Path.home() / ".bonsait"

######################
# model configuration
######################

DEFAULT_MODEL = "all-mpnet-base-v2"

######################
# base dimension table configuration
######################

BONSAI_ACTIVITY_API = "https://lca.aau.dk/api/activity-names"
BONSAI_API_KEY = os.environ.get("BONSAI_API_KEY")



import logging
from typing import Iterable, List, Optional

import torch

from bonsait.class_repo import BaseClass, EmbeddingCache
from bonsait.configs import BONSAI_ACTIVITY_API, DEFAULT_MODEL
from bonsait.model_repo import Encoder
from bonsait.utils.similarity_func import calc_cosine_similarity


class BonsaiTransformer:
    def __init__(
        self,
        source_class: str | None = None,
        target_class: BaseClass | None = None,
        model: Optional[Encoder] = None,
        device: str = "cpu",
    ) -> None:
        self._source_class = source_class
        self._target_class = target_class
        if target_class is None:
            self._target_class = BaseClass.from_bonsai(class_name="activity")
            print(
                f"get BONSAI activity classification as the default target classification from {BONSAI_ACTIVITY_API}"
            )

        if model is None:
            logging.info(
                f"No model is provided, the default model {DEFAULT_MODEL} is used"
            )
            model = Encoder.from_sentence_transformer(model_name=DEFAULT_MODEL)
        self._model = model
        self._device = device

    @property
    def target_class(self) -> Optional[List[str]]:
        return self._target_class

    @target_class.setter
    def target_class(self, value: Iterable[str]):
        if not isinstance(value, Iterable):
            raise ValueError("target_class must be a list of strings.")
        if not all(isinstance(item, str) for item in value):
            raise ValueError("Each item in target_class must be a string.")
        if not value:
            raise ValueError("target_class list cannot be empty.")
        self._target_class = value

    @property
    def source_class(self) -> Optional[str]:
        return self._source_class

    @source_class.setter
    def source_class(self, value: str):
        self._source_class = value

    def encode_source_class(self):
        array_source = self._model.encode(self._source_class).unsqueeze(0)
        return array_source

    def encode_target_class(self):
        if not self._target_class:
            raise ValueError("target_class is not set")

        encoded_class = EmbeddingCache().load_embedding(
            class_value=self._target_class.values
        )
        if encoded_class:
            print("Using cached classifications")
            return encoded_class
        else:
            print(f"Start encoding {self._target_class.name}")
            encoded_classes = [
                self._model.encode(classification)
                for classification in self._target_class.values
            ]
            EmbeddingCache().save_embedding(
                encoding=encoded_class, class_value=self._target_class.values
            )
            return encoded_classes

    def transform(
        self,
        similarity_func: callable = None,
    ):
        """
        Computes the correspondence classification from target_class
        that is most similar to the source_class based on cosine similarity.
        """

        source_vector = self.encode_source_class()
        target_vectors = self.encode_target_class()

        # Stack all target vectors to create the target matrix
        target_matrix = torch.stack(target_vectors).to(self._device)
        if similarity_func is None:
            logging.info(
                f"No similarity func provided, using the default cosine similarity: {calc_cosine_similarity.__name__}"
            )
            similarity_func = calc_cosine_similarity
        similarity_scores = similarity_func(source_vector, target_matrix)
        idx_most_similar = torch.argmax(similarity_scores, dim=1).item()

        return self._target_class.values[idx_most_similar]



from __future__ import annotations

import hashlib
import json
import os
from pathlib import Path
from typing import Iterable, Optional

import numpy as np
import requests

from bonsait.configs import BONSAI_ACTIVITY_API, BONSAI_API_KEY, CACHE_DIR


class EmbeddingCache:
    def __init__(self, cache_dir: Path | None = None) -> None:
        if not cache_dir:
            print(f"Use default cache directory {CACHE_DIR}")
            cache_dir = CACHE_DIR
        self.cache_dir = cache_dir
        os.makedirs(self.cache_dir, exist_ok=True)

    def _get_hash(self, class_value: Iterable) -> str:
        class_value_byte = json.dumps(class_value, sort_keys=True).encode(
            "utf-8"
        )  # NOTE: `sort_keys` make sure unsorted iterable get the same hash value
        return hashlib.sha256(class_value_byte).hexdigest()

    def _get_file_path(self, hash: str) -> Path:
        return self.cache_dir / f"{hash}.npy"

    def save_embedding(self, encoding, class_value: Iterable):
        hash = self._get_hash(class_value)
        file_path = self._get_file_path(hash)
        np.save(file_path, encoding)

    def load_embedding(self, class_value: Optional[Iterable]):
        if not class_value:
            return None
        hash = self._get_hash(class_value)
        file_path = self._get_file_path(hash)
        if os.path.exists(file_path):
            return np.load(file_path, allow_pickle=True)


class BaseClass:
    def __init__(self, name: str, src: Iterable) -> None:
        self.values = src
        self.name = name

    @classmethod
    def from_bonsai(
        cls,
        class_name: str,  # , cache: Optional[ClassCache] = None
    ) -> "BaseClass":
        if class_name == "activity":
            key = BONSAI_API_KEY if BONSAI_API_KEY else None
            class_activity = get_bonsai_activity_classification(key=key)
            return cls(name="activity", src=class_activity)


def get_bonsai_activity_classification(
    url: str = BONSAI_ACTIVITY_API, key: str = None
) -> Iterable[str]:
    """Get BONSAI's activity classification using its API

    Parameters
    ----------
    url : str, optional
        url for bonsai activity classification, by default BONSAI_ACTIVITY_API

    Returns
    -------
    Iterable[str]
        a list of activity classifications
    """

    try:
        headers = None
        if key:
            headers = {"Authorization": f"Token {key}"}

        response = requests.get(url, headers=headers)
        response.raise_for_status()
        activities_data = response.json()

        activity_names = [activity["description"] for activity in activities_data]
        print(f"successfully fetched activity classifications from {url}")
        return activity_names
    except requests.exceptions.HTTPError as http_err:
        print(f"HTTP error occurred: {http_err}")
    except requests.exceptions.ConnectionError as conn_err:
        print(f"Error Connecting: {conn_err}")
    except requests.exceptions.Timeout as timeout_err:
        print(f"Timeout Error: {timeout_err}")
    except requests.exceptions.RequestException as req_err:
        print(f"Error: {req_err}")  # Ambiguous error
    except Exception as err:
        print(f"An error occurred: {err}")  # Other errors

    return []



import torch
from sentence_transformers import SentenceTransformer
from transformers import BertModel, BertTokenizer


class Encoder:
    def __init__(self, encoder=None, tokenizer=None, device: str = "cpu") -> None:
        self.encoder = encoder
        self.tokenizer = tokenizer
        self.device = device
        if encoder:
            self.encoder = self.encoder.to(self.device)

    @classmethod
    def from_sentence_transformer(cls, model_name: str, device: str = "cpu"):
        encoder = SentenceTransformer(model_name)
        return cls(encoder, device=device)

    @classmethod
    def from_hugging_face(cls, model_name: str, device: str = "cpu"):
        tokenizer = BertTokenizer.from_pretrained(model_name)
        encoder = BertModel.from_pretrained(model_name)
        return cls(encoder, tokenizer, device)

    def encode(self, sentences, return_tensors="pt"):
        if self.tokenizer:  # Assuming this means we're using Hugging Face BERT
            tokens = self.tokenizer(
                sentences, padding=True, truncation=True, return_tensors=return_tensors
            )
            tokens = {key: value.to(self.device) for key, value in tokens.items()}

            with torch.no_grad():
                outputs = self.encoder(**tokens)
            embeddings = outputs.last_hidden_state[
                :, 0, :
            ]  # Use the [CLS] token embeddings
            return embeddings

        else:  # Assuming this is for SentenceTransformer
            return self.encoder.encode(sentences, convert_to_tensor=True).to(
                self.device
            )
